// SPDX-FileCopyrightText: 2023 Tobias Fella <tobias.fella@kde.org>
// SPDX-License-Identifier: GPL-2.0-or-later

#pragma once

/**
 * Initlalize logging to file and enables some additional categories, which will only be logged to the file
 */
void initLogging();
